import pytest
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis as LDA_sklearn
from scalib.modeling import LDAClassifier
import numpy as np
import scipy.stats


def test_lda():
    np.set_printoptions(threshold=np.inf)  # for debug
    ns = 10
    n_components = 2
    nc = 4
    n = 5000

    m = np.random.randint(0, 4, (nc, ns))
    traces = np.random.randint(0, 10, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    traces += m[labels]

    lda = LDAClassifier(nc, n_components, ns)
    lda.fit(traces, labels)

    lda_ref = LDA_sklearn(solver="eigen")
    lda_ref.fit(traces, labels)

    assert np.allclose(lda.lda.get_projection(), lda_ref.scalings_[:, :n_components])

    # check equivalence for means and cov in subspace
    traces_t = (lda_ref.scalings_[:, :n_components].T @ traces.T).T
    means_check = np.zeros((nc, n_components))
    for i in range(nc):
        I = np.where(labels == i)[0]
        means_check[i, :] = np.mean(traces_t[I, :], axis=0)
    traces_t = traces_t - means_check[labels, :]
    cov_check = np.cov(traces_t.T)
    assert np.allclose(means_check, lda.lda.get_means().T)
    #    assert(np.allclose(cov_check,lda.lda.get_cov()))

    traces = np.random.randint(0, 10, (n, ns), dtype=np.int16)
    labels = np.random.randint(0, nc, n, dtype=np.uint16)
    traces += m[labels]

    prs = lda.predict_proba(traces)

    traces_t = (lda_ref.scalings_[:, :n_components].T @ traces.T).T
    prs_ref = np.zeros((len(traces), nc))
    for x in range(nc):
        prs_ref[:, x] = scipy.stats.multivariate_normal.pdf(
            traces_t, means_check[x], cov_check
        )
    prs_ref = (prs_ref.T / np.sum(prs_ref, axis=1)).T

    assert np.allclose(prs, prs_ref, rtol=1e-2)



import numpy as np
from scalib import _scalib_ext


class LDAClassifier:
    r"""Models the leakage :math:`\mathbf{l}` with :math:`n_s` dimensions using
    linear discriminant analysis dimentionality reduction and gaussian
    templates.
    Based on the training data, linear discriminant analysis build a linear
    dimentionality reduction to :math:`p` dimensions that maximizes class
    separation.
    Then, a multivariate gaussian template is fitted for each class (using the
    same covariance matrix for all the classes) in the reduced dimensionality
    space to predict leakage likelihood [1]_.

    Let :math:`\mathbf{W}` be the dimensionality reduction matrix of size
    (:math:`p`, :math:`n_s`). The likelihood is

    .. math::
            \mathsf{\hat{f}}(\mathbf{l} | x) =
                \frac{1}{\sqrt{(2\pi)^{p} \cdot |\mathbf{\Sigma} |}} \cdot
                \exp^{\frac{1}{2}
                    (\mathbf{W} \cdot \mathbf{l} - \mathbf{\mu}_x)
                    \mathbf{\Sigma}
                    ( \mathbf{W} \cdot \mathbf{l}-\mathbf{\mu}_x)'}

    where :math:`\mathbf{\mu}_x` is the mean of the leakage for class :math:`x` in
    the projected space (:math:`\mu_x = \mathbb{E}(\mathbf{W}\mathbf{l}_x)`, where
    :math:`\mathbf{l}_x` denotes the leakage traces of class :math:`x`) and
    :math:`\mathbf{\Sigma}` its covariance (:math:`\mathbf{\Sigma} =
    \mathbb{Cov}(\mathbf{W}\mathbf{l}_x - \mathbf{\mu}_x)`).

    `LDAClassifier` provides the probability of each class with `predict_proba`
    thanks to Bayes' law such that

    .. math::
        \hat{\mathsf{pr}}(x|\mathbf{l}) = \frac{\hat{\mathsf{f}}(\mathbf{l}|x)}
                    {\sum_{x^*=0}^{n_c-1} \hat{\mathsf{f}}(\mathbf{l}|x^*)}.

    Examples
    --------
    >>> from scalib.modeling import LDAClassifier
    >>> import numpy as np
    >>> x = np.random.randint(0,256,(5000,10),dtype=np.int16)
    >>> y = np.random.randint(0,256,5000,dtype=np.uint16)
    >>> lda = LDAClassifier(256,3,10)
    >>> lda.fit(x,y)
    >>> x = np.random.randint(0,256,(20,10),dtype=np.int16)
    >>> predicted_proba = lda.predict_proba(x)

    Notes
    -----
    This implementation uses custom implementation of
    `sklearn.LDA(solver="eigen")` to compute the projection matrix and a custom
    implementation of `scipy.stats.multivariate_normal.pdf()`.

    .. [1] François-Xavier Standaert and Cédric Archambeau, "Using
       Subspace-Based Template Attacks to Compare and Combine Power and
       Electromagnetic Information Leakages", CHES 2008: 411-425

    Parameters
    ----------
    nc : int
        Number of possible classes (e.g., 256 for 8-bit target). `nc` must
        be smaller than `2**16`.
    p : int
        Number of dimensions in the linear subspace.
    ns: int
        Number of dimensions in the leakage.
    """

    def __init__(self, nc, p, ns):
        self.p_ = p
        self.nc_ = nc
        self.ns_ = ns
        self.lda = _scalib_ext.LDA(nc, p, ns)
        assert p < nc

    def fit(self, l, x):
        r"""Estimates the PDF parameters that is the projection matrix
        :math:`\mathbf{W}`, the means :math:`\mathbf{\mu}_x` and the covariance
        :math:`\mathbf{\Sigma}`.


        Parameters
        ----------
        l : array_like, int16
            Array that contains the traces. The array must
            be of dimension `(n,ns)` and its type must be `int16`.
        x : array_like, uint16
            Labels for each trace. Must be of shape `(n)` and
            must be `uint16`.

        Notes
        -----
        This method does not support updating the model: calling this method
        twice overrides the previous result.
        """
        self.lda.fit(l, x)

    def predict_proba(self, l):
        r"""Computes the probability for each of the classes for the traces
        contained in `l`.

        Parameters
        ----------
        l : array_like, int16
            Array that contains the traces. The array must
            be of dimension `(n,ns)` and its type must be `int16`.

        Returns
        -------
        array_like, f64
            Probabilities. Shape `(n, nc)`.
        """
        prs = self.lda.predict_proba(l)
        return prs

    def __getstate__(self):
        lda = self.lda
        dic = {
            "means": lda.get_means(),
            "cov": lda.get_cov(),
            "projection": lda.get_projection(),
            "psd": lda.get_psd(),
            "nc": self.nc_,
            "p": self.p_,
            "ns": self.ns_,
        }
        return dic

    def __setstate__(self, state):
        self.lda = _scalib_ext.LDA(state["nc"], state["p"], state["ns"])
        self.lda.set_state(
            state["cov"],
            state["psd"],
            state["means"],
            state["projection"],
            state["nc"],
            state["p"],
            state["ns"],
        )
        self.nc_ = state["nc"]
        self.ns_ = state["ns"]
        self.p_ = state["p"]



from scalib.metrics import SNR
from scalib.modeling import LDAClassifier
from scalib.attacks import SASCAGraph
from scalib.postprocessing import rank_accuracy

from utils import sbox, gen_traces
import numpy as np

if __name__ == "__main__":
    nc = 256
    npoi = 2

    # Parameters
    std = 1
    ntraces_a = 1000
    ntraces_p = 20000

    print("1. Generate simulated traces with parameters:")
    print(f"    ntraces for profiling: {ntraces_p}")
    print(f"    ntraces for attack:    {ntraces_a}")
    print(f"    noise std:             {std}")
    traces_p, labels_p = gen_traces(ntraces_p, std, random_key=1)
    traces_a, labels_a = gen_traces(ntraces_a, std, random_key=0)

    _, ns = traces_p.shape
    print("2. Profiling")
    # For each of the 16 variables at the output of the Sbox (xi)
    # we build a model that store in the dictonnary models.
    #
    # The profiling is done in 3 phases
    # 1. Compute SNR for all xi
    # 2. Select the Point-of-Interest based on the SNR
    # 3. Fit an LDA model for each of the xi

    models = {}
    for i in range(16):
        models[f"x{i}"] = {}

    print("    2.1 Compute snr for 16 Sbox outputs xi")
    # y array with xi values
    x = np.zeros((ntraces_p, 16), dtype=np.uint16)
    for i in range(16):
        x[:, i] = labels_p[f"x{i}"]

    # estimate SNR
    snr = SNR(nc=256, ns=ns, np=16)
    snr.fit_u(traces_p, x)
    snr_val = snr.get_snr()

    print(f"    2.2 Keep {npoi} POIs for each xi")
    # POI are the npoi indexes with the largest SNR
    for i in range(16):
        models[f"x{i}"]["poi"] = np.argsort(snr_val[i])[-npoi:]

    print("    2.3 Build LDAClassifier for each xi")
    for i in range(16):
        model = models[f"x{i}"]
        lda = LDAClassifier(nc=256, ns=npoi, p=1)
        lda.fit(l=traces_p[:, model["poi"]], x=labels_p[f"x{i}"].astype(np.uint16))
        model["lda"] = lda

    print("3. Attack")
    print("    3.1 Create the SASCA Graph")
    graph_desc = f"""
                NC {nc}
                TABLE sbox   # The Sbox
                """

    for i in range(16):
        graph_desc += f"""
                VAR SINGLE k{i} # The key
                VAR MULTI p{i}  # The plaintext
                VAR MULTI x{i}  # Sbox output
                VAR MULTI y{i}  # Sbox input
                PROPERTY y{i} = k{i} ^ p{i} # Key addition
                PROPERTY x{i} = sbox[y{i}]  # Sbox lookup
                """

    # Init SASCAGraph with graph_desc and the number of attack traces
    sasca = SASCAGraph(graph_desc, n=ntraces_a)

    print("    3.2 Add public information in the SASCAGraph")
    sasca.set_table("sbox", sbox)
    for i in range(16):
        sasca.set_public(f"p{i}", labels_a[f"p{i}"].astype(np.uint32))

    print("    3.3 Add xi distributions in the graph")
    for i in range(16):
        model = models[f"x{i}"]
        sasca.set_init_distribution(
            f"x{i}", model["lda"].predict_proba(traces_a[:, model["poi"]])
        )

    print("    3.4 Run belief propagation")
    print("")
    sasca.run_bp(it=3)
    print("")

    print("4. Attack evaluation")
    print("    4.1 Byte-wise attack")

    # correct secret key
    secret_key = []
    # distribution for each of the key bytes
    key_distribution = []
    # the best key guess of the adversary
    guess_key = []
    # rank for all the key bytes
    ranks = []

    for i in range(16):
        sk = labels_a[f"k{i}"][0]  # secret key byte
        distribution = sasca.get_distribution(f"k{i}")[0, :]

        guess_key.append(np.argmax(distribution))
        ranks.append(256 - np.where(np.argsort(distribution) == sk)[0])

        secret_key.append(sk)
        key_distribution.append(distribution)

    print("")
    print("        secret key (hex):", " ".join(["%3x" % (x) for x in secret_key]))
    print("        best key   (hex):", " ".join(["%3x" % (x) for x in guess_key]))
    print("        key byte ranks  :", " ".join(["%3d" % (x) for x in ranks]))
    print("")

    print(f"    4.2 Estimate full log2 key rank:")
    key_distribution = np.array(key_distribution)

    rmin, r, rmax = rank_accuracy(-np.log10(key_distribution), secret_key, 1)

    lrmin, lr, lrmax = (np.log2(rmin), np.log2(r), np.log2(rmax))
    print("")
    print(f"        {lrmin} < {lr} < {lrmax}")
    print("")



r"""Estimation of the rank of a true key given likelihood of sub-keys.

Rank estimation estimates the rank of the full (e.g. 128-bit) key based on a
score for each value of its (e.g 8-bit) sub-keys. The scores must be additive
and positive (e.g. negative log probabilities).

`rank_accuracy` allows to specify the desired precision of the bound in
bits (may not be achieved if computationally untractable), while `rank_nbin`
gives direct, lower-level control over the number of bins in the historgrams.


Notes
-----
The rank estimation algorithm is based on [1]_, with the following
optimization: computation of histogram bins with higher score than the expected
key is skipped, since it has no impact on the final rank.

.. [1] "Simple Key Enumeration (and Rank Estimation) Using Histograms: An
   Integrated Approach", R. Poussier, F.-X. Standaert, V. Grosso in CHES2016.
"""

import math

from scalib import _scalib_ext


def rank_nbin(costs, key, nbins, method="hist"):
    r"""Estimate the rank of the full keys based on scores based on histograms.

    Parameters
    ----------
    costs : array_like, f64
        Cost for each of the sub-keys. Array must be of shape `(ns,nc)` where
        `ns` is the number of sub-keys, `nc` the possible values of each
        sub-keys.
    key : array_like, int
        Correct full key split in sub-keys. Array must be of shape `(ns,)`.
    nbins : int
        Number of bins for each of the distributions.
    method : string
        Method used to estimate the rank. Can be the following:

        * "hist": using histograms (default).
        * "ntl": using NTL library, allows better precision.

    Returns
    -------
    (rmin, r, rmax): (float, float, float)

            - **rmin** is a lower bound for the key rank.
            - **r** is the stimated key rank.
            - **rmax** is an upper bound for the key rank.
    """
    return _scalib_ext.rank_nbin(costs, key, nbins, _choose_merge_value(costs), method)


def rank_accuracy(costs, key, acc_bit=1.0, method="hist"):
    r"""Estimate the rank of the full keys based on scores based on histograms.

    Parameters
    ----------
    costs : array_like, f64
        Cost for each of the sub-keys. Array must be of shape `(ns,nc)` where
        `ns` is the number of sub-keys, `nc` the possible values of each
        sub-keys.
    key : array_like, int
        Correct full key split in sub-keys. Array must be of shape `(ns,)`.
    acc_bit : f64
        Expected log2 accuracy for the key rank estimation. `acc_bit` must so be
        set to `log2(rank_max/rank_min)`.
    method : string
        Method used to estimate the rank. Can be the following:

        * "hist": using histograms (default).
        * "ntl": using NTL library, allows better precision.

    Returns
    -------
    (rmin, r, rmax): (float, float, float)

            - **rmin** is a lower bound for the key rank.
            - **r** is the stimated key rank.
            - **rmax** is an upper bound for the key rank.
    """

    return _scalib_ext.rank_accuracy(
        costs, key, 2.0 ** acc_bit, _choose_merge_value(costs), method
    )


def _choose_merge_value(costs):
    """The merge parameter is the number of sub-keys to merge in a
    brute-force manner before computing histograms. Merging may improve
    accuracy at the expense of running time.
    Here we limit sub-histograms to 2**16 values.
    """
    return max(1, int(16 / math.log2(max(len(c) for c in costs))))



import numpy as np
from scalib import _scalib_ext


class SNR:
    r"""Computes the Signal-to-Noise Ratio (SNR) between the traces and the
    intermediate values. Informally, SNR allows to quantify the amount of information about a
    random variable :math:`X` contained in the mean of the leakage :math:`L_X`. High SNR
    means more information contained in the mean. The SNR metric is defined
    with the following equation [1]_:

    .. math::
        \mathrm{SNR} = \frac{\mathrm{Var}_{x\leftarrow X}(\mathrm{E}[L_x])}
                {\mathrm{E}_{x\leftarrow X}(\mathrm{Var}[L_x])}

    The SNR is estimated from leakage samples :math:`L_x` and the value `x` of
    the random variable.

    Parameters
    ----------
    nc : int
        Number of possible values for the random variable :math:`X` (e.g., 256 for 8-bit
        target). `nc` must be between :math:`1` and :math:`2^{16}` (included).
    ns : int
        Number of samples in a single trace.
    np : int
        Number of independent variables `X` for which SNR must be estimated.

    Examples
    --------
    >>> from scalib.metrics import SNR
    >>> import numpy as np
    >>> traces = np.random.randint(0,256,(100,200),dtype=np.int16)
    >>> X = np.random.randint(0,256,(100,10),dtype=np.uint16)
    >>> snr = SNR(256,200,10)
    >>> snr.fit_u(traces,X)
    >>> snr_val = snr.get_snr()

    Notes
    -----
    .. [1] "Hardware Countermeasures against DPA ? A Statistical Analysis of
       Their Effectiveness", Stefan Mangard, CT-RSA 2004: 222-235

    """

    def __init__(self, nc, ns, np=1):
        if nc not in range(1, 2 ** 16 + 1):
            raise ValueError(
                f"SNR can be computed on max 16 bit variable, nc={nc} given"
            )

        self._ns = ns
        self._np = np
        self._snr = _scalib_ext.SNR(nc, ns, np)

    def fit_u(self, l, x):
        r"""Updates the SNR estimation with samples of `l` for the classes `x`.
        This method may be called multiple times.

        Parameters
        ----------
        l : array_like, np.int16
            Array that contains the signal. The array must
            be of dimension `(n, ns)` and its type must be `np.int16`.
        x : array_like, np.uint16
            Labels for each trace. Must be of shape `(n, np)` and must be
            `np.uint16`.
        """
        nl, nsl = l.shape
        nx, npx = x.shape
        if not (npx == self._np and nx == nl):
            raise ValueError(f"Expected x with shape ({nl}, {self._np})")
        if not (nsl == self._ns):
            raise Exception(f"l is too long. Expected second dim of size {self._ns}.")
        # _scalib_ext uses inverted axes for x.
        self._snr.update(l, x.transpose())

    def get_snr(self):
        r"""Return the current SNR estimation with an array of shape `(np,ns)`."""
        return self._snr.get_snr()




# Based on https://scikit-hep.org/developer/packaging

from setuptools import setup
from setuptools_rust import Binding, RustExtension

# Ensure these are present (in case we are not using PEP-518 compatible build
# system).
import setuptools_scm
import toml


setup(
    project_urls={
        "Bug Tracker": "https://github.com/simple-crypto/scalib/issues",
    },
    rust_extensions=[
        RustExtension(
            "scalib._scalib_ext",
            path="src/scalib_ext/scalib-py/Cargo.toml",
            binding=Binding.PyO3,
            features=["pyo3/abi3"],
            py_limited_api=True,
        )
    ],
)
